"""
    file        sc_mqtt_client.py
    brief       SpringCore - SpringPark demonstration
    author      Matthieu Barreteau <matthieu.b@springcard.com>
    date        27/01/2021
 
    copyright   Copyright (c) 2021 SPRINGCARD SAS, FRANCE - www.springcard.com
    version     MBA 27/01/2021: initial release, plain connexion

    Requirements:
    - Python >= 3.7.3
    - Paho MQTT library (python -m pip install paho-mqtt)

    Purpose:
    The aim of this project is to demonstrate how to interact
    with a SpringPark reader using MQTT.
    The reader publish a message using MQTT each time it reads a card.
    This event is then received by this demonstration client.
    The demonstation client, in return and if it accepts the card, 
    publish an answer for the reader with a UI command (LED and Buzzer).

    You can stop this script using CTRL+C
    """

# libraries import
from threading import Event
import paho.mqtt.client as mqtt
import logging
import getopt
import json
import time
import ssl
import sys
import os
import re
import binascii

# helpers import
from helpers.mqtt_events import mqtt_events
from helpers.ctrlccatcher import CtrlCCatcher


# Incoming message callback
def incoming_message( client, userdata, msg ):

    # sanity check
    if ( client == None ) or ( msg == None ):
        return

    logging.info( f"New message on {msg.topic}: {str(msg.payload)}" )

    # try load this json content
    try:
        json_content = json.loads( msg.payload )
    except json.JSONDecodeError:
        logging.warning( "This is not a json message!" )
        return
    
    # json message is correct, check if we can find the TagId field
    if ( "valid_uids" in userdata ) and ( "TagId" in json_content ):

        """ this is a demo command to ask the reader to play some 
            error sound! """
        message = '{ "Sequence": "24"}'

        # change for a command topic
        topic = re.sub( "/evt$", "/cmd", msg.topic )

        # check if the tag is present in our "valid tags list"
        tag = json_content[ "TagId" ]
        if tag in userdata[ "valid_uids" ]:
            """ this tag's UID is in our list, let's send a message
            to open the door """
            logging.info( f"Tag {tag} is valid!" )

            """ this is a demo command to ask the reader to play some sound
            and open a virtual door """
            message = '{ "Sequence": "23", "Relay": true }'
        else:
            logging.info( f"Tag {tag} is invalid!" )

        logging.info( f"Sending answer on {topic}: {str(message)}" )
        client.publish( topic=topic, payload=message )

# display_help function
def display_help( script_name ):

    print( script_name )
    print( "-s, --server=<server>           MQTT server" )
    print( "-p, --port=<port>               MQTT port" )
    print( "[--nolog]                       Silent mode" )
    print( "[--logtofile]                   Output to debug.log" )
    print( "[--ca=<file>]                   CA certificate file" )
    print( "[--cert=<file>]                 Client's certificate file" )
    print( "[--key=<file>]                  Client's private key file" )
    print( "[--aws]                         Use AWS protocol options" )
    print( "[--clientid]                    MQTT Client ID" )
    sys.exit()


# main function  
def main( script_name, argv ):

    # custom settings for AC demo
    current_path = os.path.dirname( os.path.realpath( __file__ ) )
    server = "mqtt.springcard.com"
    port = 8883
    keepalive = 60
    must_exit = False
    use_log = 1
    mqtt_context = {}
    client_id = "TEST-PYTHON-"+binascii.b2a_hex(os.urandom(4)).decode('ascii').upper()

    # SSL options
    aws_mode = True
    use_ssl = True
    ca = current_path+"/../CA/springcore-cloud-ca-chain.crt"
    cert = current_path+"/../Clients/watch/test-watch.crt"
    key = current_path+"/../Clients/watch/test-watch.key"

    mqtt_context[ "current_path" ] = current_path

    # options parsing
    try:
        opts, _ = getopt.getopt(argv,"hs:p:", [ "help", "server=","port=", "nolog", "logtofile", \
                                        "ca=", "cert=", "key=", "aws", "clientid=" ])
    except getopt.GetoptError as e:
        print(e)
        display_help( script_name )

    for opt, arg in opts:
        if opt in ( "-h", "--help"):
            display_help( script_name )
        elif opt in ( "-s", "--server"):
            server = arg
        elif opt == "--nolog":
            use_log = 0
        elif opt == "--logtofile":
            use_log = 2        
        elif opt == "--ca":
            ca = arg
        elif opt == "--cert":
            cert = arg
        elif opt == "--key":
            key = arg
        elif opt == "--aws":
            aws_mode = True
        elif opt == "--clientid":
            client_id = arg            
        elif opt in ( "-p", "--port"):
            try:
                port = int(arg)
            except ValueError:
                print( "Port must be an integer number!" )
                sys.exit()

    # set debug options
    log_level = logging.DEBUG 
    if use_log:
        if use_log > 1:
            logging.basicConfig(format='%(asctime)s - %(levelname)s - %(message)s', \
                                datefmt='%d-%b-%y %H:%M:%S', \
                                filename=f"debug.log", filemode="w", \
                                level=log_level)
        else:
            logging.basicConfig(format='%(asctime)s - %(levelname)s - %(message)s', \
                                datefmt='%d-%b-%y %H:%M:%S', \
                                level=log_level)    


    # try to setup SSL context
    if ( ca != "" ) and ( cert != "" ) and ( key != "" ):
        try:
            #debug print opnessl version
            logging.info( f"using {ssl.OPENSSL_VERSION}" )
            ssl_context = ssl.create_default_context()
            if aws_mode:
                ssl_context.set_alpn_protocols( [ "x-amzn-mqtt-ca" ] )
                mqtt_context[ "aws_mode" ] = None
            ssl_context.load_verify_locations( cafile=ca )
            ssl_context.load_cert_chain( certfile=cert, keyfile=key )
            use_ssl = True
        except Exception:
            print( "Could not create SSL context!" )
            sys.exit()


    # CTRL+C catcher thread (this could be useful on some Windows platforms)
    ctrlccatch = CtrlCCatcher()
    ctrlccatch.start()

    # process exit flag
    mqtt_context[ "exit_event" ] = Event()
    mqtt_context[ "exit_event" ].clear()    

    # callback declaration
    mqtt_context[ "callback" ] = incoming_message

    # select on which topics to subscribe (topic, qos)
    mqtt_context[ "subscriptions" ] = [
        ( "springcard/springcore/+/rdr/evt", 0 ),
    ]
    
    # list of tag's UID we accept for this demonstration
    mqtt_context[ "valid_uids" ] = [
        "0477BB02C84080", 
        "0084F42B",
        "04EB7A429E4880",
        "0496DEA2784D80",
        "046B2DA2794280",
        "04A17B02C84080"
    ]


    # start our MQTT client process
    client = mqtt.Client( client_id=client_id )
    if use_ssl:
        client.tls_set_context( context=ssl_context )
    client.user_data_set( mqtt_context )
    client.on_connect = mqtt_events.on_connect
    client.on_disconnect = mqtt_events.on_disconnect
    client.on_message = mqtt_events.on_message

    logging.info( f"Trying to connect to {server} on port {port}" )
    try:
        client.connect( host=server, port=port, keepalive=keepalive )
    except ( TimeoutError, ConnectionRefusedError ):
        logging.error( "Unable to contact this server!" )
        must_exit = True

    if not must_exit:
        # start mqtt processing thread
        client.loop_start()

        # this is the main loop, it does nothing but wait for a quit indication
        while ctrlccatch.wait() and not mqtt_context[ "exit_event" ].is_set():
            # no need to eat all cpus
            time.sleep(0.5)

        # stop the client
        client.loop_stop()


    # stop all process and exit
    logging.info( "Cleaning..." )
    ctrlccatch.terminate()
    ctrlccatch.join()
    logging.info( "Done." )

  
# entry point
if __name__ == "__main__":
    # add "this script" folder to the library search path
    # could be useful on Windows platform
    sys.path.append( os.path.dirname( os.path.realpath( __file__ ) ) )
  
    # start our application
    main( os.path.basename( __file__ ), sys.argv[1:] )
    
# EOF   